/**
 * A topological layout using {@link GridOperator}.
 *
 * @module
 */
import { Dag } from "../dag";
import { LaneOperator } from "./lane";
import { GreedyOperator } from "./lane/greedy";
declare type OpDag<Lop extends LaneOperator> = Lop extends LaneOperator<infer N, infer L> ? Dag<N, L> : never;
/**
 * The return from calling {@link GridOperator}
 *
 * This is the final width and height of the laid out dag.
 */
export interface GridInfo {
    width: number;
    height: number;
}
/**
 * A simple grid based topological layout operator.
 *
 * This layout algorithm constructs a topological representation of the dag
 * meant for visualization. The nodes are topologically ordered and then nodes
 * are put into lanes such that an edge can travel horizontally to the lane of
 * a child node, and then down without intersecting to that child.
 *
 * This layout produces good representations when you want a compressed layout
 * where each node gets its only line formation.
 *
 * Create with {@link grid}.
 *
 * @example
 * <img alt="grid example" src="media://grid-greedy-topdown.png" width="200">
 *
 * @example
 * ```typescript
 * const data = [["parent", "child"], ...];
 * const create = connect();
 * const dag = create(data);
 * const layout = grid();
 * const { width, height } = layout(dag);
 * for (const node of dag) {
 *   console.log(node.x, node.y);
 * }
 * ```
 */
export interface GridOperator<Lane extends LaneOperator> {
    /** Layout the input dag */
    (dag: OpDag<Lane>): GridInfo;
    /**
     * Set the lane operator to the given {@link LaneOperator} and returns a new
     * version of this operator. (default: {@link GreedyOperator})
     */
    lane<NewLane extends LaneOperator>(val: NewLane): GridOperator<NewLane>;
    /** Get the current lane operator */
    lane(): Lane;
    /**
     * Sets this grid layout's node size to the specified two-element array of
     * numbers [ *width*, *height* ] and returns a new operator. These sizes are
     * effectively the grid size, e.g. the spacing between adjacent lanes or rows
     * in the grid. (default: [1, 1])
     *
     * Note, due to the way that edges are meant to rendered, edges won't
     * intersect with nodes if width is half of the actual node width.
     */
    nodeSize(val: readonly [number, number]): GridOperator<Lane>;
    /** Get the current node size. */
    nodeSize(): [number, number];
    /**
     * Update the grid layout size
     *
     * Sets this grid layout's node size to the specified two-element array of
     * numbers [ *width*, *height* ] and returns a new operator. After the
     * initial layout, if size is not null, the dag will be rescaled so that it
     * fits in width x height. (default: null)
     */
    size(val: null | readonly [number, number]): GridOperator<Lane>;
    /** Get the current size. */
    size(): null | [number, number];
}
export declare type DefaultGridOperator = GridOperator<GreedyOperator>;
/**
 * Create a new {@link GridOperator} with default settings.
 */
export declare function grid(...args: never[]): DefaultGridOperator;
export {};
